/* $Id: list-int.c,v 1.14 2010/03/05 01:23:31 oohara Exp $ */

#include <stdio.h>
/* malloc */
#include <stdlib.h>

#include "list-int.h"

list_int *
list_int_new(int n)
{
  list_int *p = NULL;

  p = (list_int *) malloc(sizeof(list_int));
  if (p == NULL)
  {
    fprintf(stderr, "list_int_new: malloc failed\n");
    return NULL;
  }

  p->prev = NULL;
  p->next = NULL;
  p->n = n;

  return p;
}

list_int *
list_int_unlink(list_int *p)
{
  list_int *q = NULL;

  if (p == NULL)
    return NULL;

  q = NULL;
  if (p->next != NULL)
  {
    q = p->next;
    p->next->prev = p->prev;
  }
  if (p->prev != NULL)
  {
    q = p->prev;
    p->prev->next = p->next;
  }

  p->next = NULL;
  p->prev = NULL;

  return q;
}

void
list_int_delete(list_int *p)
{
  if (p == NULL)
    return;

  list_int_unlink(p);
  free(p);
  p = NULL;
}

list_int *
list_int_first(list_int *p)
{
  list_int *q = NULL;

  if (p == NULL)
    return NULL;

  q = p;
  while (q->prev != NULL)
  {
    if (q->prev == p)
    {
      fprintf(stderr, "list_int_first: circular list found\n");
      return NULL;
    }
    q = q->prev;
  }
  
  return q;
}

list_int *
list_int_last(list_int *p)
{
  list_int *q = NULL;

  if (p == NULL)
    return NULL;

  q = p;
  while (q->next != NULL)
  {
    if (q->next == p)
    {
      fprintf(stderr, "list_int_last: circular list found\n");
      return NULL;
    }
    q = q->next;
  }
  
  return q;
}

void
list_int_delete_all(list_int *p)
{
  list_int *q = NULL;
  list_int *q_temp = NULL;

  if (p == NULL)
    return;

  q = list_int_first(p);
  while (q != NULL)
  {
    q_temp = q->next;
    list_int_delete(q);
    q = q_temp;
  }
}

list_int *
list_int_concat(list_int *p, list_int *q)
{
  list_int *p_last = NULL;
  list_int *q_first = NULL;

  if (p == NULL)
    return q;
  if (q == NULL)
    return p;

  p_last = list_int_last(p);
  q_first = list_int_first(q);
  if (p_last == NULL)
  {
    fprintf(stderr, "list_int_concat: p_last is NULL\n");
    return NULL;
  }
  if (q_first == NULL)
  {
    fprintf(stderr, "list_int_concat: q_first is NULL\n");
    return NULL;
  }

  p_last->next = q_first;
  q_first->prev = p_last;

  return p;
}

list_int *
list_int_insert_after(list_int *where, list_int *p)
{
  list_int *p_first = NULL;
  list_int *p_last = NULL;

  if (where == NULL)
    return p;
  if (p == NULL)
    return where;

  p_first = list_int_first(p);
  p_last = list_int_last(p);
  if (p_first == NULL)
  {
    fprintf(stderr, "list_int_insert_after: p_first is NULL\n");
    return NULL;
  }
  if (p_last == NULL)
  {
    fprintf(stderr, "list_int_insert_after: p_last is NULL\n");
    return NULL;
  }

  if (where->next != NULL)
  {  
    where->next->prev = p_last;
    p_last->next = where->next;
  }
  where->next = p_first;
  p_first->prev = where;

  return where;
}

list_int *
list_int_insert_before(list_int *where, list_int *p)
{
  list_int *p_first = NULL;
  list_int *p_last = NULL;

  if (where == NULL)
    return p;
  if (p == NULL)
    return where;

  p_first = list_int_first(p);
  p_last = list_int_last(p);
  if (p_first == NULL)
  {
    fprintf(stderr, "list_int_insert_before: p_first is NULL\n");
    return NULL;
  }
  if (p_last == NULL)
  {
    fprintf(stderr, "list_int_insert_before: p_last is NULL\n");
    return NULL;
  }

  if (where->prev != NULL)
  {  
    where->prev->next = p_first;
    p_first->prev = where->prev;
  }
  where->prev = p_last;
  p_last->next = where;

  return p;
}

list_int *
list_int_append(list_int *p, int n)
{
  return list_int_concat(p, list_int_new(n));
}

list_int *
list_int_prepend(list_int *p, int n)
{
  return list_int_concat(list_int_new(n), p);
}

list_int *
list_int_merge(list_int *p, list_int *q, void *data,
               int (*compare)(list_int *, list_int *, void *))
{
  list_int *temp_p = NULL;
  list_int *temp_q = NULL;
  list_int *result = NULL;

  temp_p = list_int_first(p);
  temp_q = list_int_first(q);
  while ((temp_p != NULL) || (temp_q != NULL))
  {
    if (temp_p == NULL)
    {
      result = list_int_concat(result, temp_q);
      return result;
    }
    if (temp_q == NULL)
    {
      result = list_int_concat(result, temp_p);
      return result;
    }

    if (((*compare)(temp_p, temp_q, data)) <= 0)
    {
      result = list_int_concat(result, temp_p);
      temp_p = temp_p->next;
      if (temp_p != NULL)
      {
        temp_p->prev->next = NULL;
        temp_p->prev = NULL;
      }
    }
    else
    {
      result = list_int_concat(result, temp_q);
      temp_q = temp_q->next;
      if (temp_q != NULL)
      {
        temp_q->prev->next = NULL;
        temp_q->prev = NULL;
      }
    }
  }

  return result;
}

list_int *
list_int_sort(list_int *p, void *data,
              int (*compare)(list_int *, list_int *, void *))
{
  list_int *p_first = NULL;
  list_int *p_half = NULL;
  list_int *p2 = NULL;

  p_first = list_int_first(p);
  if ((p_first == NULL) || (p_first->next == NULL))
    return p_first;

  p_half = p_first;
  p2 = p_first;
  while (p2 != NULL)
  {
    p_half = p_half->next;
    p2 = p2->next;
    if (p2 != NULL)
      p2 = p2->next;
  }
  p_half->prev->next = NULL;
  p_half->prev = NULL;

  return list_int_merge(list_int_sort(p_first, data, compare),
                        list_int_sort(p_half, data, compare),
                        data, compare);
}

int
list_int_length_all(list_int *p)
{
  int n;
  list_int *q = NULL;

  n = 0;
  q = list_int_first(p);
  while (q != NULL)
  {
    n++;
    q = q->next;
  }

  return n;
}

